"""Code for processing strings with brackets

This module contains functions for processing and simplifying strings
containing brackets.  It provides functionality for identifying whether
brackets in a string form matched pairs (every opening bracket has
a corresponding closing bracket, in the correct order), among other
capabilities.
"""

from typing import Union

from pyxx.strings.exceptions import (
    NotABracketError,
    UnmatchedBracketsError,
)


def _check_valid_brackets(opening_bracket: str, closing_bracket: str,
                          opening_bracket_name: str = 'opening_bracket',
                          closing_bracket_name: str = 'closing_bracket'
                          ) -> bool:
    """Checks that a given set of opening and closing brackets form a
    valid pair

    Verifies that a given pair of opening and closing brackets are a
    valid choice for the PyXX bracket-matching algorithm.  To meet this
    criterion, brackets must be strings of a single character in length,
    and the opening and closing bracket must be different.

    Parameters
    ----------
    opening_bracket : str
        Character representing opening bracket
    closing_bracket : str
        Character representing closing bracket
    opening_bracket_name : str, optional
        Name that should be used to refer to the ``opening_bracket`` argument
        in error messages (default is ``'opening_bracket'``)
    closing_bracket_name : str, optional
        Name that should be used to refer to the ``closing_bracket`` argument
        in error messages (default is ``'closing_bracket'``)

    Returns
    -------
    bool
        Whether ``opening_bracket`` and ``closing_bracket`` form a valid pair
        of brackets

    Raises
    ------
    TypeError
        If either ``opening_bracket`` or ``closing_bracket`` are not of
        type "str"
    ValueError
        If either ``opening_bracket`` or ``closing_bracket`` don't have length
        1, or if ``opening_bracket`` and ``closing_bracket`` are identical
    """
    if not (isinstance(opening_bracket, str)
            and isinstance(closing_bracket, str)):
        raise TypeError(f'Arguments "{opening_bracket_name}" and '
                        f'"{closing_bracket_name}" must be of type "str"')

    if (len(opening_bracket) != 1) or (len(closing_bracket) != 1):
        raise ValueError(f'Arguments "{opening_bracket_name}" and '
                         f'"{closing_bracket_name}" must have length 1')

    if opening_bracket == closing_bracket:
        raise ValueError(f'Arguments "{opening_bracket_name}" and '
                         f'"{closing_bracket_name}" must be different '
                         'characters')

    return True


def contains_all_matched_brackets(value: str, opening_bracket: str = '(',
                                  closing_bracket: str = ')') -> bool:
    """Checks whether all opening brackets in a string have a
    corresponding closing bracket

    Parameters
    ----------
    value : str
        String to process
    opening_bracket : str, optional
        Character representing opening bracket (default is ``'('``)
    closing_bracket : str, optional
        Character representing closing bracket (default is ``')'``)

    Returns
    -------
    bool
        Returns ``True`` if for every opening bracket there is a
        single corresponding closing bracket that occurs later in
        the string, and ``False`` otherwise
    """
    # Validate inputs
    if not (isinstance(value, str)):
        raise TypeError('Argument "value" must be of type "str"')

    _check_valid_brackets(opening_bracket, closing_bracket)

    # Parse string, checking for matched brackets
    counter = 0
    for char in value:
        if char == opening_bracket:
            counter += 1
        elif char == closing_bracket:
            counter -= 1

        # If counter is negative, this indicates that closing bracket
        # occurred before the corresponding opening bracket
        if counter < 0:
            return False

    return counter == 0


def find_matching_bracket(value: str, begin: int, opening_bracket: str = '(',
                          closing_bracket: str = ')') -> int:
    """Finds the index of the bracket that forms a matched pair with another
    bracket at a given index

    Finds the "other bracket" that forms a closed pair of matched brackets in
    a string ``value``, beginning from the bracket at index ``begin``. Note
    that if the character at index ``begin`` is an opening bracket, then
    ``value`` is searched in the forward direction (left to right); while if
    it is a closing bracket, then the search direction is reversed (right to
    left).

    Parameters
    ----------
    value : str
        String to search for matching brackets
    begin : int
        Index in ``value`` where one of the brackets in the matched
        pair is found
    opening_bracket : str, optional
        Character representing opening bracket (default is ``'('``)
    closing_bracket : str, optional
        Character representing closing bracket (default is ``')'``)

    Returns
    -------
    int
        Returns the index of the matching bracket, or ``-1`` if the
        bracket at index ``begin`` in ``value`` does not have a
        matching bracket in ``value``

    Raises
    ------
    ValueError
        If the character at index ``begin`` in ``value`` is not one of:
        ``opening_bracket`` or ``closing_bracket``
    """
    # Adjust if user provides an index relative to the end of the string
    if begin < 0:
        begin += len(value)

    # Define search direction and ending index depending on whether the
    # `begin` index specifies an opening or closing bracket
    begin_char = value[begin]

    if begin_char == opening_bracket:
        k = 1
        end = len(value)
    elif begin_char == closing_bracket:
        k = -1
        end = -1
    else:
        raise NotABracketError(
            f'Character at index {begin} of "{value}" is "{begin_char}", '
            'which is not an opening or closing bracket')

    # Search string to find the matching bracket
    counter = 0
    for i in range(begin, end, k):
        if value[i] == opening_bracket:
            counter += 1
        elif value[i] == closing_bracket:
            counter -= 1

        # The first time `counter` returns to zero, the matching
        # bracket has been found
        if counter == 0:
            return i

    return -1


def find_skip_brackets(value: str, target_chars: Union[str, tuple],
                       begin: int, direction: str = 'forward',
                       opening_bracket: str = '(', closing_bracket: str = ')'
                       ) -> int:
    """Finds the index of a target character, skipping over matched brackets

    This function is useful for finding the index of a particular character
    in a string, ignoring content inside of matched brackets.  It begins
    at a given index in a string and searches until any of a given set of
    target character(s) are found.  If an opening or closing bracket is
    reached, the function skips to the corresponding matched bracket before
    resuming the search (i.e., content inside of matched brackets is not
    searched).

    Parameters
    ----------
    value : str
        String to search
    target_chars : str or tuple
        A string or tuple with one or more characters to search for
    begin : int
        Index in ``value`` at which to begin searching
    direction : str, optional
        Whether to search ``value`` in order of increasing or decreasing
        index.  Can be either ``'forward'`` or ``'reverse'`` (default
        is ``'forward'``)
    opening_bracket : str, optional
        Character representing opening bracket (default is ``'('``)
    closing_bracket : str, optional
        Character representing closing bracket (default is ``')'``)

    Returns
    -------
    int
        The index of the first occurrence of any characters in ``target_chars``
        found in ``value``, beginning the search at index ``begin`` and
        searching in the direction specified by ``direction`` and ignoring
        content inside of matched brackets.  If none of ``target_chars``
        are found, ``-1`` is returned

    Notes
    -----
    When searching, the function will not "enter" matched brackets; however,
    if the index specified by ``begin`` is inside a set of brackets, the
    function will search inside of these brackets and can "leave" these
    brackets and search in parts of the string outside these brackets
    """
    # Verify that "value" argument is a string
    if not isinstance(value, str):
        raise TypeError('Argument "value" must be of type "str"')

    # Verify that "begin" is a valid index for the given string
    if not (-len(value) <= begin <= len(value) - 1):
        raise IndexError(
            f'Index {begin} at which to begin search is not valid for '
            f'string "{value}" (length {len(value)})')

    # Adjust if user provides an index relative to the end of the string
    i = begin + len(value) if begin < 0 else begin

    # Set increment by which to advance to next character when
    # searching string
    if direction not in ('forward', 'reverse'):
        raise ValueError('Argument "direction" must be one of: '
                         '"forward" or "reverse"')

    k = 1 if direction == 'forward' else -1

    # Verify that "value" argument doesn't have unmatched brackets
    if not contains_all_matched_brackets(value, opening_bracket,
                                         closing_bracket):
        raise UnmatchedBracketsError(
            'Argument "value" contains unmatched brackets')

    while 0 <= i < len(value):
        # Check whether target character has been found
        if value[i] in target_chars:
            return i

        # Check whether an opening or closing bracket has been
        # found; if so, move to the end of the matched pair
        if any(((direction == 'forward' and value[i] == opening_bracket),
                (direction == 'reverse' and value[i] == closing_bracket))):

            if (i := find_matching_bracket(
                    value, i, opening_bracket, closing_bracket)) == -1:

                # This error should never be thrown since the string was
                # previously check for matching brackets; however, it's
                # included as an added precaution
                raise UnmatchedBracketsError(   # pragma: no cover
                    f'Unmatched brackets found in string "{value}"')

        # Advance to next character in string
        i += k

    return -1


def strip_matched_brackets(value: str, max_pairs: int = -1, strip: bool = True,
                           return_num_pairs_removed: bool = False,
                           opening_bracket: str = '(', closing_bracket: str = ')'):
    """Remove matched leading/trailing brackets from strings

    Removes matched brackets from a string (i.e., if a string begins with
    ``opening_bracket`` and the corresponding closing bracket
    ``closing_bracket`` is the last character in the string, the brackets
    are removed), as well as (optionally) any leading and/or trailing
    whitespace.

    Parameters
    ----------
    value : str
        String from which to remove matched brackets and leading/trailing
        whitespace (if ``strip`` is ``True``)
    max_pairs : int, optional
        Maximum pairs of matched brackets to remove.  Set to ``-1`` to
        remove an unlimited number of matched brackets (default is ``-1``)
    strip : bool, optional
        Whether to remove leading and/or trailing whitespace when processing
        ``value`` (default is ``True``)
    return_num_pairs_removed : bool, optional
        Whether to return the number of pairs of brackets that were removed
        (default is ``False``)

    Returns
    -------
    str
        Input string ``value`` with matched brackets removed
    int
        The number of pairs of matched brackets that were removed (returned
        if and only if ``return_num_pairs_removed`` is ``True``)

    Notes
    -----
    If ``strip`` is set to ``False``, then matched brackets are removed if and
    only if the first character in ``value`` is ``opening_bracket`` and the
    last character is ``closing_bracket`` and they form a matching pair
    """
    value = value.strip() if strip else value

    num_removed = 0
    while (value.startswith(opening_bracket) and value.endswith(closing_bracket)):
        # If the maximum number of pairs of brackets have been
        # removed, exit the loop
        if num_removed >= max_pairs >= 0:
            break

        # Only remove leading and trailing brackets if they form
        # a matching pair
        if find_matching_bracket(value, -1,
                                 opening_bracket=opening_bracket,
                                 closing_bracket=closing_bracket) == 0:
            value = value[1:-1].strip() if strip else value[1:-1]
            num_removed += 1
        else:
            break

    # Output results
    if not return_num_pairs_removed:
        return value

    return value, num_removed



"""Classes for controlling loop execution.
"""

import signal


class InterruptibleLoop:
    """Context manager to create an interruptible loop

    This context manager allows users to interrupt and terminate a loop early
    by pressing ``Ctrl+c``.  This can be useful to stop a long process early
    while monitoring results (such as stopping training of a machine learning
    model).  When interrupted, the context manager will change its
    :py:attr:`interrupted` attribute to ``True``.  Additionally, if the
    ``throw_exception`` option is enabled, an ``InterruptedError`` will be
    raised (which can be caught and handled as desired).

    Examples
    --------
    Basic usage:

    >>> with pyxx.dev.InterruptibleLoop() as loop:
    ...     for i in range(1000):
    ...         # Do something
    ...
    ...         if loop.interrupted:
    ...             print('Loop was interrupted')
    ...             break

    Example of using the ``throw_exception`` option:

    >>> try:
    ...     with pyxx.dev.InterruptibleLoop(throw_exception=True) as loop:
    ...         for i in range(1000):
    ...             pass  # Do something
    ...
    ... except InterruptedError:
    ...     print('Loop was interrupted')
    ...
    ... else:
    ...     print('Loop was NOT interrupted')
    Loop was NOT interrupted
    """

    def __init__(self, throw_exception: bool = False) -> None:
        # Store user inputs
        self.__throw_exception = throw_exception

        # Initialize context manager internal state
        self.__interrupted = False
        self.__original_sigint_handler = signal.getsignal(signal.SIGINT)

    @property
    def interrupted(self) -> bool:
        """Whether the loop has been interrupted"""
        return self.__interrupted

    def _interrupt_handler(self, *args, **kwargs) -> None:  # pylint: disable=W0613
        self.__interrupted = True

        if self.__throw_exception:
            raise InterruptedError

    def __enter__(self) -> 'InterruptibleLoop':
        signal.signal(signal.SIGINT, self._interrupt_handler)
        self.__interrupted = False

        return self

    def __exit__(self, *args, **kwargs) -> None:
        signal.signal(signal.SIGINT, self.__original_sigint_handler)



"""This module contains functions for parsing strings containing units and
providing a dictionary of each unit contained in the string and the exponent
to which the unit is raised.  For instance, the unit ``'kg*m/s^2'`` should
be converted to the format: ``{'kg': 1.0, 's': -2.0, 'm': 1.0}``.
"""

import copy
import string

# Mypy exclusion added for SymPy since it is not typed
import sympy  # type: ignore

from pyxx.strings.functions.brackets import (
    contains_all_matched_brackets,
    find_skip_brackets,
    strip_matched_brackets,
)
from pyxx.strings.functions.content import (
    str_excludes_chars,
    str_includes_only,
)
from pyxx.strings.functions.split import split_at_index
from pyxx.units.exceptions import (
    InvalidUnitError,
    InvalidExponentError,
    ParserMaxIterationError,
)


def _add_to_dict(dictionary: dict, unit: str, exponent: float):
    """Supporting function for :py:func:`parse_unit` that adds a unit
    to the output dictionary

    Used by :py:func:`parse_unit` to add a unit to the output dictionary,
    combining units with the same key by adding exponents (since these
    units are multiplied, so they can be described equivalently by a unit
    with the sum of their exponents).

    Parameters
    ----------
    dictionary : dict
        Output dictionary to which to add units
    unit : str
        Key of unit to add to ``dictionary``
    exponent : float
        Exponent corresponding to ``key``
    """
    if unit in dictionary:
        dictionary[unit] += exponent
    else:
        dictionary[unit] = exponent


def _check_dict_keys(dictionary: dict, special_chars: str):
    """Verifies that all keys in a dictionary do not have certain characters

    Checks each key in a given input dictionary to see whether it contains
    prohibited characters and returns the result.  This is useful when parsing
    unit strings, since iterating until this function returns ``True`` ensures
    that all compound units have been simplified.

    Parameters
    ----------
    dictionary : dict
        Dictionary whose keys are to be analyzed
    special_chars : str
        Characters which, if found in any dictionary keys, should result in
        the function returning ``False``

    Returns
    -------
    bool
        Returns ``True`` if none of the characters in ``special_chars`` are
        found in the keys of ``dictionary`` and ``False`` otherwise
    """
    for key in dictionary.keys():
        if not _is_unit(key, special_chars):
            return False

    return True


def _is_unit(value: str, special_chars: str):
    """Checks whether a string is a fully-simplified unit with no compound
    unit operators

    Checks whether a string is a fully-simplified unit by analyzing whether
    it contains any user-specified special characters (such as ``*`` or ``^``)
    that are used to form compound units.  If these characters are found, the
    string is determined to not be a unit.

    Parameters
    ----------
    value : str
        String to analyze
    special_chars : str
        Characters which, if found in ``value``, indicate that ``value``
        is a compound unit

    Returns
    -------
    bool
        Returns ``True`` if ``value`` does not contain any characters in
        ``special_chars``, and ``False`` otherwise
    """
    return str_excludes_chars(value, special_chars)


def parse_unit(unit: str, max_iterations: int = 10000):
    """Parses a string of simple or compound units, separating it into
    fully-simplified component units

    Iteratively parses a string containing a set of multiplied and/or
    divided units, identifying each fully-simplified component unit and the
    power to which it is raised, and returning these data in a dictionary.
    For instance, the unit ``'kg*m/s^2'`` should be parsed and result in
    the dictionary: ``{'kg': 1.0, 's': -2.0, 'm': 1.0}``.  Standard order
    of operations are followed when parsing units (parentheses evaluated
    first, followed by exponents, then multiplication and division).

    Parameters
    ----------
    unit : str
        String containing a simple or compound unit to parse
    max_iterations : int, optional
        Iteration limit for iteratively parsing the unit (default is
        ``10000``)

    Returns
    -------
    dict
        Returns a Python dictionary.  Each key represents a fully-simplified
        component unit parsed from ``unit``, and the corresponding value is
        the exponent to which the unit is raised in ``unit`` (i.e., the
        product of all keys raised to their corresponding values is equivalent
        to the original unit string ``unit``)

    Warnings
    --------
    When parsing units, the following are considered "special characters" that
    are not allowed to be used in fully-simplified units: ``0123456789.*/^()``

    A unit is considered fully-simplified if it does not contain any of these
    special characters.  Attempting to include any of these characters in the
    simple component units may result in the compound unit being impossible
    to parse, or other unexpected behavior.

    Notes
    -----
    It is assumed that parentheses ``()`` are used to group components of
    compound units.  Square brackets ``[]`` and braces ``{}`` are not
    recognized as valid types of brackets for grouping components of units.

    Examples
    --------
    Consider a relatively simple unit, like :math:`m/s` or :math:`m/s^2`.
    Parsing these units results in dictionaries with the component units
    as keys and their exponents as the corresponding values:

    >>> pyxx.units.parse_unit('m/s')
    {'s': -1.0, 'm': 1.0}
    >>> pyxx.units.parse_unit('m/s^2')
    {'s': -2.0, 'm': 1.0}

    The main value of PyXX's unit parser is that it can handle much more
    complex cases.  For instance, units with parentheses can be parsed:

    >>> pyxx.units.parse_unit('kg*(m/s)^2')
    {'kg': 1.0, 's': -2.0, 'm': 2.0}

    Units with nested parentheses and math expressions in their exponents
    can be parsed:

    >>> pyxx.units.parse_unit('kg / (m^(1/2+3.14) / s^(5-7))^3')
    {'s': -6.0, 'm': -10.92, 'kg': 1.0}

    Any units can be parsed, as long as they don't use the prohibited
    characters specified in the "Warnings" section.

    >>> pyxx.units.parse_unit('my_unit^3 * degreesC')
    {'my_unit': 3.0, 'degreesC': 1.0}
    """
    # Store original unit string
    original_unit = unit

    # Special characters in units
    operator_chars = '*/^()'             # Used to construct compound units
    number_chars = string.digits + '.'   # Characters reserved for numbers
    special_chars = operator_chars + number_chars

    # Verify that input is a string
    if not isinstance(unit, str):
        raise TypeError('Argument "unit" must be of type "str"')

    # Verify that all parentheses in the input are matched
    if not contains_all_matched_brackets(unit):
        raise ValueError('Parentheses in argument "unit" are not all matched')

    # Replace Python exponent characters ** with ^
    unit = unit.replace('**', '^')

    # Remove whitespace
    unit = unit.replace(' ', '').replace('\t', '')

    # Initialize output data structure
    output = {} if unit == '' else {unit: 1.0}

    # Parse unit
    n = -1
    while (not (success := _check_dict_keys(output, special_chars))) \
            and (((n := n + 1) < max_iterations) or (max_iterations < 0)):

        # Iterate through each key in dictionary and simplify it
        for unit_string in copy.deepcopy(output):

            # STEP 1) Remove key from output dictionary
            exponent = output.pop(unit_string)

            # If unit string is enclosed by matched parentheses, remove
            # them, as they are unnecessary
            unit_string = strip_matched_brackets(unit_string)

            # STEP 2) If unit is does not contain any special characters, then
            # it is already simplified as much as possible, so move to next key
            if _is_unit(unit_string, special_chars):
                _add_to_dict(output, unit_string, exponent)
                continue

            # STEP 3) Separate units that are multiplied together (e.g., N*m)
            #   Work forward through the string, eliminating any units that are
            #   multiplied by each other
            _continue = False
            while (i := find_skip_brackets(unit_string, target_chars='*',
                                           begin=0, direction='forward')) != -1:
                _continue = True

                # Throw error if unit begins with "*"
                if i == 0:
                    raise InvalidUnitError(
                        f'Invalid unit string "{unit_string}" encountered '
                        f'when parsing unit "{original_unit}" (unit strings '
                        'cannot begin with "*")')

                # Split `unit_string` at the "*" character
                x1, x2 = split_at_index(unit_string, i)

                # Store the first part of the split string, and repeat loop to
                # check whether there are any other multiplied units in the
                # second part of the split string
                _add_to_dict(output, strip_matched_brackets(x1), exponent)
                unit_string = strip_matched_brackets(x2)

            if _continue:
                _add_to_dict(output, strip_matched_brackets(unit_string), exponent)
                continue

            # STEP 4) Separate units that are divided (e.g., m/s)
            #   Work backward through the string, eliminating any units that are
            #   divided by each other
            while (i := find_skip_brackets(unit_string, target_chars='/',
                                           begin=-1, direction='reverse')) != -1:
                _continue = True

                # Throw error if unit ends with "/"
                if i == len(unit_string) - 1:
                    raise InvalidUnitError(
                        f'Invalid unit string "{unit_string}" encountered '
                        f'when parsing unit "{original_unit}" (unit strings '
                        'cannot end with "/")')

                # Split `unit_string` at the "/" character
                x1, x2 = split_at_index(unit_string, i)

                # Store the second part of the split string with a negative
                # exponent, and repeat loop to check whether there are any
                # other divided units in the first part of the split string
                _add_to_dict(output, strip_matched_brackets(x2), -exponent)
                unit_string = strip_matched_brackets(x1)

            if _continue:
                _add_to_dict(output, strip_matched_brackets(unit_string), exponent)
                continue

            # STEP 5) Simplify units that are raised to a power (e.g., m^2)
            if (i := find_skip_brackets(unit_string, target_chars='^',
                                        begin=-1, direction='reverse')) != -1:
                # Throw error if unit begins or ends with "^"
                if not (1 <= i <= len(unit_string) - 2):
                    raise InvalidUnitError(
                        f'Invalid unit string "{unit_string}" encountered '
                        f'when parsing unit "{original_unit}" (unit strings '
                        'cannot begin or end with "^")')

                # Split unit string at last "^" character
                base, exp = split_at_index(unit_string, i)

                # Verify that exponent is a number
                try:
                    # For security, verify that the exponent contains only
                    # expected characters
                    strip_exp = strip_matched_brackets(exp)
                    if not str_includes_only(strip_exp, special_chars + 'e+-'):
                        raise ValueError

                    # Attempt to simplify exponent
                    exp = float(sympy.simplify(strip_exp))

                except (ValueError, TypeError) as exception:
                    raise InvalidExponentError(
                        f'Invalid exponent "{exp}" encountered when '
                        f' parsing unit "{original_unit}"') from exception

                # Verify that base is either a unit or enclosed in
                # matched parentheses
                if not (_is_unit(base, special_chars)
                        or (base != strip_matched_brackets(base))):
                    raise InvalidUnitError(
                        f'Invalid base "{base}" encountered when parsing '
                        f'unit "{original_unit}" (base must either be a unit '
                        'or enclosed in parentheses')

                # Store unit
                _add_to_dict(output, strip_matched_brackets(base), exponent*exp)
                continue

            # STEP 6) Fallback step
            #   If none of the previous steps could simplify the unit,
            #   store it back inside the unit dictionary to be processed
            #   again on the next iteration
            _add_to_dict(output, unit_string, exponent)

    if not success:
        raise ParserMaxIterationError(
            f'Unable to parse unit string "{original_unit}" after '
            f'{max_iterations} iterations')

    # Remove any units with an exponent of zero
    output = {key: value for key, value in output.items() if value != 0}

    return output



"""This module contains classes intended to represent a system of
units (that is, a system with a set of base units from which all
other relevant units can be derived).
"""

from typing import Optional


class UnitSystem:
    """Base class representing a system of units

    This class can be used to represent a system of units with an arbitrary
    number of base units.  For more information about systems of units and
    unit conversions, refer to the :ref:`section-units_concepts` section
    of the documentation.
    """

    def __init__(self, num_base_units: int, name: Optional[str] = None,
                 description: Optional[str] = None):
        """Creates an instance of the :py:class:`UnitSystem` class

        Creates a system of units, assigning the number of base units
        (which cannot be changed later) and optionally an identifying name
        and description.

        Parameters
        ----------
        num_base_units : int
            The number of base units in the system of units
        name : str, optional
            A short, descriptive name of the system of units (default is
            ``None``)
        description : str, optional
            A description of the system of units (default is ``None``)
        """
        # Store number of base units
        if not isinstance(num_base_units, int):
            raise TypeError(
                'Argument "num_base_units" must be of type "int"')
        if num_base_units <= 0:
            raise ValueError(
                'Argument "num_base_units" must be positive')
        self._num_base_units = num_base_units

        # Store name and description
        self.name = name
        self.description = description

    def __repr__(self):
        return str(self)

    def __str__(self):
        representation = str(self.__class__)

        if self.name is not None:
            representation += f' - {self.name}'

        if self.description is not None:
            representation += f' - {self.description}'

        return representation

    @property
    def description(self):
        """A brief description of the system of units, such as a list of the
        base units or details on other relevant specifications"""
        return self._description

    @description.setter
    def description(self, description: Optional[str]):
        if not (isinstance(description, str) or description is None):
            raise TypeError('Argument "description" must be of type "str"')

        self._description = description

    @property
    def name(self):
        """A short, descriptive name identifying the system of units"""
        return self._name

    @name.setter
    def name(self, name: Optional[str]):
        if not (isinstance(name, str) or name is None):
            raise TypeError('Argument "name" must be of type "str"')

        self._name = name

    @property
    def num_base_units(self):
        """The number of base units in the system of units"""
        return self._num_base_units


class UnitSystemSI(UnitSystem):
    """Class representing the SI system of units

    Class that can be used to represent the International System of
    Units (SI).  This system of units has 7 base units, which are
    described in the "Notes" section.

    Notes
    -----
    The base units in the SI system are:

    1. Length: meter [:math:`m`]
    2. Time: second [:math:`s`]
    3. Amount of substance: mole [:math:`mol`]
    4. Electric current: ampere [:math:`A`]
    5. Temperature: Kelvin [:math:`K`]
    6. Luminous intensity: candela [:math:`cd`]
    7. Mass: kilogram [:math:`kg`]

    References
    ----------
    https://www.nist.gov/pml/weights-and-measures/metric-si/si-units
    """

    def __init__(self, name: Optional[str] = 'SI',
                 description: Optional[str] = 'International System of Units'):
        """Creates an instance of the :py:class:`UnitSystemSI` class

        Creates a system of units corresponding to the International System of
        Units (SI).  This system of units has 7 base units.

        Parameters
        ----------
        name : str, optional
            A short, descriptive name of the system of units (default is
            ``'SI'``)
        description : str, optional
            A description of the system of units (default is ``'International
            System of Units'``)
        """
        # Configure the appropriate number of base units (7) for SI
        num_base_units = 7

        # Initialize the object
        super().__init__(
            num_base_units=num_base_units,
            name=name,
            description=description
        )
